#ifdef PETSC_RCS_HEADER
static char vcid[] = "$Id: ex19.c,v 1.1 1999/03/26 21:15:43 bsmith Exp bsmith $";
#endif

static char help[] ="\
  -mx <xg>, where <xg> = number of grid points in the x-direction\n\
  -my <yg>, where <yg> = number of grid points in the y-direction\n\
  -Nx <npx>, where <npx> = number of processors in the x-direction\n\
  -Ny <npy>, where <npy> = number of processors in the y-direction\n\n";

/*  
    This problem is modeled by
    the partial differential equation
  
            -Laplacian u  = g,  0 < x,y < 1 ,
  
    with boundary conditions
   
             u = 0  for  x = 0, x = 1, y = 0, y = 1.
  
    A finite difference approximation with the usual 5-point stencil
    is used to discretize the boundary value problem to obtain a nonlinear 
    system of equations.
*/

#include "sles.h"
#include "da.h"
#include "mg.h"

/* User-defined application contexts */

typedef struct {
   int        mx,my;            /* number grid points in x and y direction */
   Vec        localX,localF;    /* local vectors with ghost region */
   DA         da;
   Vec        x,b,r;            /* global vectors */
   Mat        J;                /* Jacobian on grid */
} GridCtx;

typedef struct {
   GridCtx     fine;
   GridCtx     coarse;
   SLES        sles_coarse;
   int         ratio;
   Mat         I;               /* interpolation from coarse to fine */
} AppCtx;

#define COARSE_LEVEL 0
#define FINE_LEVEL   1

extern int FormInterpolation(AppCtx *);
extern int FormJacobian_Grid(AppCtx *,GridCtx *, Mat *);

/*
      Mm_ratio - ration of grid lines between fine and coarse grids.
*/
#undef __FUNC__
#define __FUNC__ "main"
int main( int argc, char **argv )
{
  AppCtx        user;                      
  int           ierr, its, N, n, Nx = PETSC_DECIDE, Ny = PETSC_DECIDE;
  int           size, flg,nlocal,Nlocal;
  double        bratu_lambda_max = 6.81, bratu_lambda_min = 0.;
  SLES          sles,sles_fine;
  PC            pc;
  Scalar        one = 1.0;

  PetscInitialize( &argc, &argv,(char *)0,help );

  user.ratio = 2;
  user.coarse.mx = 5; user.coarse.my = 5; 
  ierr = OptionsGetInt(PETSC_NULL,"-Mx",&user.coarse.mx,&flg); CHKERRA(ierr);
  ierr = OptionsGetInt(PETSC_NULL,"-My",&user.coarse.my,&flg); CHKERRA(ierr);
  ierr = OptionsGetInt(PETSC_NULL,"-ratio",&user.ratio,&flg); CHKERRA(ierr);
  user.fine.mx = user.ratio*(user.coarse.mx-1)+1; user.fine.my = user.ratio*(user.coarse.my-1)+1;

  PetscPrintf(PETSC_COMM_WORLD,"Coarse grid size %d by %d\n",user.coarse.mx,user.coarse.my);
  PetscPrintf(PETSC_COMM_WORLD,"Fine grid size %d by %d\n",user.fine.mx,user.fine.my);

  n = user.fine.mx*user.fine.my; N = user.coarse.mx*user.coarse.my;

  MPI_Comm_size(PETSC_COMM_WORLD,&size);
  ierr = OptionsGetInt(PETSC_NULL,"-Nx",&Nx,&flg); CHKERRA(ierr);
  ierr = OptionsGetInt(PETSC_NULL,"-Ny",&Ny,&flg); CHKERRA(ierr);

  /* Set up distributed array for fine grid */
  ierr = DACreate2d(PETSC_COMM_WORLD,DA_NONPERIODIC,DA_STENCIL_STAR,user.fine.mx,
                    user.fine.my,Nx,Ny,1,1,PETSC_NULL,PETSC_NULL,&user.fine.da); CHKERRA(ierr);
  ierr = DACreateGlobalVector(user.fine.da,&user.fine.x); CHKERRA(ierr);
  ierr = VecDuplicate(user.fine.x,&user.fine.r); CHKERRA(ierr);
  ierr = VecDuplicate(user.fine.x,&user.fine.b); CHKERRA(ierr);
  ierr = VecGetLocalSize(user.fine.x,&nlocal);CHKERRA(ierr);
  ierr = DACreateLocalVector(user.fine.da,&user.fine.localX); CHKERRA(ierr);
  ierr = VecDuplicate(user.fine.localX,&user.fine.localF); CHKERRA(ierr);
  ierr = MatCreateMPIAIJ(PETSC_COMM_WORLD,nlocal,nlocal,n,n,5,PETSC_NULL,3,PETSC_NULL,&user.fine.J); CHKERRA(ierr);

  /* Set up distributed array for coarse grid */
  ierr = DACreate2d(PETSC_COMM_WORLD,DA_NONPERIODIC,DA_STENCIL_STAR,user.coarse.mx,
                    user.coarse.my,Nx,Ny,1,1,PETSC_NULL,PETSC_NULL,&user.coarse.da); CHKERRA(ierr);
  ierr = DACreateGlobalVector(user.coarse.da,&user.coarse.x); CHKERRA(ierr);
  ierr = VecDuplicate(user.coarse.x,&user.coarse.b); CHKERRA(ierr);
  ierr = VecGetLocalSize(user.coarse.x,&Nlocal);CHKERRA(ierr);
  ierr = DACreateLocalVector(user.coarse.da,&user.coarse.localX); CHKERRA(ierr);
  ierr = VecDuplicate(user.coarse.localX,&user.coarse.localF); CHKERRA(ierr);
  ierr = MatCreateMPIAIJ(PETSC_COMM_WORLD,Nlocal,Nlocal,N,N,5,PETSC_NULL,3,PETSC_NULL,&user.coarse.J); CHKERRA(ierr);

  /* Create linear solver */
  ierr = SLESCreate(PETSC_COMM_WORLD,&sles);CHKERRA(ierr);

  /* set two level additive Schwarz preconditioner */
  ierr = SLESGetPC(sles,&pc); CHKERRA(ierr);
  ierr = PCSetType(pc,PCMG); CHKERRA(ierr);
  ierr = MGSetLevels(pc,2); CHKERRA(ierr);
  ierr = MGSetType(pc,MGADDITIVE); CHKERRA(ierr);

  ierr = FormJacobian_Grid(&user,&user.coarse,&user.coarse.J);CHKERRA(ierr);
  ierr = FormJacobian_Grid(&user,&user.fine,&user.fine.J);CHKERRA(ierr);

  /* Create coarse level */
  ierr = MGGetCoarseSolve(pc,&user.sles_coarse); CHKERRA(ierr);
  ierr = SLESSetOptionsPrefix(user.sles_coarse,"coarse_"); CHKERRA(ierr);
  ierr = SLESSetFromOptions(user.sles_coarse); CHKERRA(ierr);
  ierr = SLESSetOperators(user.sles_coarse,user.coarse.J,user.coarse.J,DIFFERENT_NONZERO_PATTERN);CHKERRA(ierr);
  ierr = MGSetX(pc,COARSE_LEVEL,user.coarse.x);CHKERRA(ierr); 
  ierr = MGSetRhs(pc,COARSE_LEVEL,user.coarse.b);CHKERRA(ierr); 

  /* Create fine level */
  ierr = MGGetSmoother(pc,FINE_LEVEL,&sles_fine); CHKERRA(ierr);
  ierr = SLESSetOptionsPrefix(sles_fine,"fine_"); CHKERRA(ierr);
  ierr = SLESSetFromOptions(sles_fine); CHKERRA(ierr);
  ierr = SLESSetOperators(sles_fine,user.fine.J,user.fine.J,DIFFERENT_NONZERO_PATTERN);CHKERRA(ierr);
  ierr = MGSetR(pc,FINE_LEVEL,user.fine.r);CHKERRA(ierr); 
  ierr = MGSetResidual(pc,FINE_LEVEL,MGDefaultResidual,user.fine.J); CHKERRA(ierr);

  /* Create interpolation between the levels */
  ierr = FormInterpolation(&user);CHKERRA(ierr);
  ierr = MGSetInterpolate(pc,FINE_LEVEL,user.I);CHKERRA(ierr);
  ierr = MGSetRestriction(pc,FINE_LEVEL,user.I);CHKERRA(ierr);

  ierr = SLESSetOperators(sles,user.fine.J,user.fine.J,DIFFERENT_NONZERO_PATTERN);CHKERRA(ierr);

  ierr = VecSet(&one,user.fine.b);CHKERRQ(ierr);
  {
    PetscRandom rand;
    ierr = PetscRandomCreate(PETSC_COMM_WORLD,RANDOM_DEFAULT,&rand);CHKERRA(ierr);
    ierr = VecSetRandom(rand,user.fine.b);CHKERRA(ierr);
    ierr = PetscRandomDestroy(rand);CHKERRA(ierr);
  }

  /* Set options, then solve nonlinear system */
  ierr = SLESSetFromOptions(sles); CHKERRA(ierr);

  ierr = SLESSolve(sles,user.fine.b,user.fine.x,&its); CHKERRA(ierr);
  PetscPrintf(PETSC_COMM_WORLD,"Number of iterations = %d\n", its );

  /* Free data structures */
  ierr = MatDestroy(user.fine.J); CHKERRA(ierr);
  ierr = VecDestroy(user.fine.x); CHKERRA(ierr);
  ierr = VecDestroy(user.fine.r); CHKERRA(ierr);
  ierr = VecDestroy(user.fine.b); CHKERRA(ierr);
  ierr = DADestroy(user.fine.da); CHKERRA(ierr);
  ierr = VecDestroy(user.fine.localX); CHKERRA(ierr);
  ierr = VecDestroy(user.fine.localF); CHKERRA(ierr);

  ierr = MatDestroy(user.coarse.J); CHKERRA(ierr);
  ierr = VecDestroy(user.coarse.x); CHKERRA(ierr);
  ierr = VecDestroy(user.coarse.b); CHKERRA(ierr);
  ierr = DADestroy(user.coarse.da); CHKERRA(ierr);
  ierr = VecDestroy(user.coarse.localX); CHKERRA(ierr);
  ierr = VecDestroy(user.coarse.localF); CHKERRA(ierr);

  ierr = SLESDestroy(sles); CHKERRA(ierr);
  ierr = MatDestroy(user.I);CHKERRA(ierr); 
  PetscFinalize();

  return 0;
}

#undef __FUNC__
#define __FUNC__ "FormJacobian_Grid"
int FormJacobian_Grid(AppCtx *user,GridCtx *grid, Mat *J)
{
  Mat     jac = *J;
  int     ierr, i, j, row, mx, my, xs, ys, xm, ym, Xs, Ys, Xm, Ym, col[5];
  int     nloc, *ltog, grow;
  Scalar  two = 2.0, one = 1.0, v[5], hx, hy, hxdhy, hydhx, sc, *x, value;
  Vec     localX = grid->localX;

  mx = grid->mx;            my = grid->my;            
  hx = one/(double)(mx-1);  hy = one/(double)(my-1);
  sc = hx*hy;               hxdhy = hx/hy;            hydhx = hy/hx;

  /* Get ghost points */
  ierr = DAGetCorners(grid->da,&xs,&ys,0,&xm,&ym,0); CHKERRQ(ierr);
  ierr = DAGetGhostCorners(grid->da,&Xs,&Ys,0,&Xm,&Ym,0); CHKERRQ(ierr);
  ierr = DAGetGlobalIndices(grid->da,&nloc,&ltog); CHKERRQ(ierr);

  /* Evaluate Jacobian of function */
  for (j=ys; j<ys+ym; j++) {
    row = (j - Ys)*Xm + xs - Xs - 1; 
    for (i=xs; i<xs+xm; i++) {
      row++;
      grow = ltog[row];
      if (i > 0 && i < mx-1 && j > 0 && j < my-1) {
        v[0] = -hxdhy; col[0] = ltog[row - Xm];
        v[1] = -hydhx; col[1] = ltog[row - 1];
        v[2] = two*(hydhx + hxdhy); col[2] = grow;
        v[3] = -hydhx; col[3] = ltog[row + 1];
        v[4] = -hxdhy; col[4] = ltog[row + Xm];
        ierr = MatSetValues(jac,1,&grow,5,col,v,INSERT_VALUES); CHKERRQ(ierr);
      } else if ((i > 0 && i < mx-1) || (j > 0 && j < my-1)){
        value = .5*two*(hydhx + hxdhy);
        ierr = MatSetValues(jac,1,&grow,1,&grow,&value,INSERT_VALUES); CHKERRQ(ierr);
      } else {
        value = .25*two*(hydhx + hxdhy);
        ierr = MatSetValues(jac,1,&grow,1,&grow,&value,INSERT_VALUES); CHKERRQ(ierr);
      }
    }
  }
  ierr = MatAssemblyBegin(jac,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
  ierr = MatAssemblyEnd(jac,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);

  return 0;
}


/* 
   Unfortunately the PETSc multigrid interface expects the interpolation to 
   be provided as a multtrans() and the restriction as a mult() so we need
   to provide a shell matrix that reverses the roles
*/
typedef struct {
  Vec vec;
  Mat mat;
} Mat_Ours;

/* this is our interpolation add routine */
int MatMultTransAdd_Ours(Mat shell,Vec x, Vec y, Vec z)
{
  Mat mat;
  int ierr;

  ierr = MatShellGetContext(shell,(void **)&mat);CHKERRQ(ierr);
  ierr = MatMultAdd(mat,x,y,z);CHKERRQ(ierr);
  return 0;
}
/* this is our restriction routine */
int MatMult_Ours(Mat shell,Vec x, Vec y)
{
  Mat mat;
  int ierr;

  ierr = MatShellGetContext(shell,(void **)&mat);CHKERRQ(ierr);
  ierr = MatMultTrans(mat,x,y);CHKERRQ(ierr);
  return 0;
}
int MatDestroy_Ours(Mat shell)
{
  Mat mat;
  int ierr;

  ierr = MatShellGetContext(shell,(void **)&mat);CHKERRQ(ierr);
  ierr = MatDestroy(mat);CHKERRQ(ierr);
  return 0;
}


#undef __FUNC__
#define __FUNC__ "FormInterpolation"
/*
      Forms the interpolation (and restriction) operator from 
coarse grid to fine.
*/
int FormInterpolation(AppCtx *user)
{
  int      ierr,i,j,i_start,m_fine,j_start,m,n,M,Mx = user->coarse.mx,My = user->coarse.my,*idx;
  int      m_ghost,n_ghost,*idx_c,m_ghost_c,n_ghost_c,m_coarse;
  int      row,i_start_ghost,j_start_ghost,cols[4],mx = user->fine.mx, m_c,my = user->fine.my;
  int      c0,c1,c2,c3,nc,ratio = user->ratio,i_end,i_end_ghost,m_c_local,m_fine_local;
  int      i_c,j_c,i_start_c,j_start_c,n_c,i_start_ghost_c,j_start_ghost_c;
  Scalar   v[4],x,y;
  Mat      mat;
  
  ierr = DAGetCorners(user->fine.da,&i_start,&j_start,0,&m,&n,0);CHKERRQ(ierr);
  ierr = DAGetGhostCorners(user->fine.da,&i_start_ghost,&j_start_ghost,0,&m_ghost,&n_ghost,0);CHKERRQ(ierr);
  ierr = DAGetGlobalIndices(user->fine.da,PETSC_NULL,&idx); CHKERRQ(ierr);

  ierr = DAGetCorners(user->coarse.da,&i_start_c,&j_start_c,0,&m_c,&n_c,0);CHKERRQ(ierr);
  ierr = DAGetGhostCorners(user->coarse.da,&i_start_ghost_c,&j_start_ghost_c,0,&m_ghost_c,&n_ghost_c,0);CHKERRQ(ierr);
  ierr = DAGetGlobalIndices(user->coarse.da,PETSC_NULL,&idx_c); CHKERRQ(ierr);

  /* create interpolation matrix */
  ierr = VecGetLocalSize(user->fine.x,&m_fine_local);CHKERRQ(ierr);
  ierr = VecGetLocalSize(user->coarse.x,&m_c_local);CHKERRQ(ierr);
  ierr = VecGetSize(user->fine.x,&m_fine);CHKERRQ(ierr);
  ierr = VecGetSize(user->coarse.x,&m_coarse);CHKERRQ(ierr);
  ierr = MatCreateMPIAIJ(PETSC_COMM_WORLD,m_fine_local,m_c_local,m_fine,m_coarse,
                         5,0,3,0,&mat);CHKERRQ(ierr);

  /* loop over local fine grid nodes setting interpolation for those*/
  for ( j=j_start; j<j_start+n; j++ ) {
    for ( i=i_start; i<i_start+m; i++ ) {
      row    = idx[m_ghost*(j-j_start_ghost) + (i-i_start_ghost)];

      i_c = (i/ratio);    /* coarse grid node to left of fine grid node */
      j_c = (j/ratio);    /* coarse grid node below fine grid node */

      /* 
         Only include those interpolation points that are truly 
         nonzero. Note this is very important for final grid lines
         in x and y directions; since they have no right/top neighbors
      */
      x  = ((double)(i - i_c*ratio))/((double)ratio);
      y  = ((double)(j - j_c*ratio))/((double)ratio);
      /* printf("i j %d %d %g %g\n",i,j,x,y); */
      nc = 0;
      /* one left and below; or we are right on it */
      cols[nc] = Mx*j_c + i_c;
      v[nc++]  = x*y - x - y + 1.0;
      /* one right and below */
      if (i_c*ratio != i) { 
        cols[nc] = Mx*j_c + i_c + 1;
        v[nc++]  = -x*y + x;
      }
      /* one left and above */
      if (j_c*ratio != j) { 
        cols[nc] = Mx*j_c + i_c + Mx;
        v[nc++]  = -x*y + y;
      }
      /* one right and above */
      if (j_c*ratio != j && i_c*ratio != i) { 
        cols[nc] = Mx*j_c + i_c + Mx + 1;
        v[nc++]  = x*y;
      }
      ierr = MatSetValues(mat,1,&row,nc,cols,v,INSERT_VALUES); CHKERRQ(ierr); 
    }
  }
  ierr = MatAssemblyBegin(mat,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
  ierr = MatAssemblyEnd(mat,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);

  ierr = MatCreateShell(PETSC_COMM_WORLD,m_c_local,m_fine_local,m_coarse,m_fine,
                        mat,&user->I);CHKERRA(ierr);
  ierr = MatShellSetOperation(user->I,MATOP_MULT,(void*)MatMult_Ours);CHKERRA(ierr);
  ierr = MatShellSetOperation(user->I,MATOP_MULT_TRANS_ADD,(void*)MatMultTransAdd_Ours);CHKERRA(ierr);
  ierr = MatShellSetOperation(user->I,MATOP_DESTROY,(void*)MatDestroy_Ours);CHKERRA(ierr);
  return 0;
}




